import logging
import logging.config
import yaml

DEFAULT_LOGGING_CFG_NAME = 'logging.yaml'

def load(targetCfgPath: str = DEFAULT_LOGGING_CFG_NAME) -> dict:
    # Load the logging config file
    with open(targetCfgPath, 'rt') as yaml_file:
        logging_cfg_map = yaml.safe_load(yaml_file.read())
    
    # init logging
    logging.config.dictConfig(logging_cfg_map)

    # return dictionary
    return logging_cfg_map



DEFAULT_USER_AGENT_CHROME = r'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/121.0.0.0 Safari/537.36'

CFG_KEY_BROWSER_BINARY = 'mw_pyhelper.webbot.browserBinary'
CFG_KEY_BROWSER_TYPE = 'mw_pyhelper.webbot.browserType'
CFG_KEY_HEADLESS_MODE = 'mw_pyhelper.webbot.headlessMode'
CFG_KEY_EXPLICIT_WAIT_TIMEOUT = 'mw_pyhelper.webbot.explicitWaitTimeout'


from .webdriver_mgr import *

# wire driver
from seleniumwire import webdriver as wire_webdriver

from . import constants

def get_webdriver(appcfg: AppCfg) -> Union[wire_webdriver.Chrome, wire_webdriver.Firefox, None]:
   mylogger = logging.getLogger(__name__)

   is_headless = False
   if appcfg.get(constants.CFG_KEY_HEADLESS_MODE) and appcfg.get(constants.CFG_KEY_HEADLESS_MODE).lower() == 'true':
      is_headless = True

   cfg_value = appcfg.get(constants.CFG_KEY_BROWSER_TYPE)
   if cfg_value and cfg_value.lower() == 'chrome':
      chrome_option = wire_webdriver.ChromeOptions()
      
      if appcfg.get(constants.CFG_KEY_BROWSER_BINARY):
         chrome_option.binary_location = appcfg.get(constants.CFG_KEY_BROWSER_BINARY)

      if is_headless:
         chrome_option.add_argument('--headless=new')
         #override user-agent as some site disallow headless mode browser
         chrome_option.add_argument('--user-agent=' + constants.DEFAULT_USER_AGENT_CHROME)
      
      mylogger.debug('Chrome type driver return')
      return wire_webdriver.Chrome(options=chrome_option)
   elif cfg_value and cfg_value.lower() == 'firefox':
      firefox_option = wire_webdriver.FirefoxOptions()
      if appcfg.get(constants.CFG_KEY_BROWSER_BINARY):
         firefox_option.binary_location = appcfg.get(constants.CFG_KEY_BROWSER_BINARY)
      
      if is_headless:
         firefox_option.add_argument('-headless')

      mylogger.debug('Firefox type driver return')
      return wire_webdriver.Firefox(options=firefox_option)
   
   mylogger.debug('No webdriver')
   return None


from typing import Dict, Optional
import os
import logging
from jproperties import Properties

# Combine cfgloader and logcfg
from .logcfg import DEFAULT_LOGGING_CFG_NAME
from .logcfg import load as load_logcfg

DEFAULT_APP_CFG_NAME = 'app.properties'


class AppCfg:
    def __init__(self, targetCfgPath: str = DEFAULT_APP_CFG_NAME) -> None:
        # Define internal logger
        self.mylogger = logging.getLogger(__name__)
        # Load config into map
        self.propertyMap = self.__load(targetCfgPath)


    def __load(self, targetCfgPath: str) -> Dict[str, str]:
        self.mylogger.debug('Loading appCfg file: %s' % targetCfgPath)

        cfgMap: Dict[str, str] = {}
        properties_tuple = Properties()

        # if cfg file not exist, return empty dictionary
        if not os.path.isfile(targetCfgPath):
            self.mylogger.info('appCfg not exist !!! path: %s' % targetCfgPath)
            return cfgMap

        # Load the app config file
        with open(targetCfgPath, 'rb') as cfg_file:
            properties_tuple.load(cfg_file)

        # Migrate tuple into dictionary
        for key in properties_tuple.keys():
            cfgMap[key] = str(properties_tuple[key].data)

        return cfgMap
    
    def debug_content(self) -> None:
        self.mylogger.debug('Listing appProperties ...')

        # Just to list out the content to debug log
        for cfg_tuple in self.propertyMap.items():
            self.mylogger.debug('%s : %s' % (cfg_tuple[0], cfg_tuple[1]))

    def get(self, key: str, include_env_variables = True, default_value: Optional[str] = None) -> Optional[str]:
        if include_env_variables:
            #We need replace all '.' value with '_' for environment variables
            env_var = os.environ.get(key.replace('.', '_'))

            #If environment variable exist, just return it (as expected env override cfg value)
            if env_var:
                return env_var
            
        #If no env_var, then just get from cfg
        return_value = self.propertyMap.get(key)
        if return_value is None and default_value is not None:
            return default_value
        return return_value
    
    def remove(self, key: str) -> Optional[str]:
        if key in self.propertyMap:
            return self.propertyMap.pop(key)
        
        return None
    
    


import logging
import requests
from typing import Optional

URL_IPIFY_API_IPV4 = r'https://api.ipify.org?format=json'
URL_IPIFY_API_IPV6 = r'https://api6.ipify.org?format=json'

def get_ip_addr(is_ipv4: bool = True) -> Optional[str]:
    mylogger = logging.getLogger(__name__)

    api_url = URL_IPIFY_API_IPV4 if is_ipv4 else URL_IPIFY_API_IPV6

    mylogger.debug(f'Calling ip-api service url: {api_url}')
    response = requests.get(api_url)
    if response.status_code != requests.codes.ok:
        mylogger.debug(f'get_ip_addr error with {api_url}: {response.text}')
        return None
    
    json_response = response.json()
    return json_response['ip']



import logging
import re
from typing import Optional, List, Union, Pattern

from playwright.sync_api import Playwright, Browser, Page, Locator, expect

from ..cfgloader import AppCfg
from . import constants

def get_browser(appcfg: AppCfg, playwright: Playwright) -> Optional[Browser]:
    if playwright is None:
        mylogger.debug('No playwright')
        return None

    mylogger = logging.getLogger(__name__)

    is_headless = False
    if appcfg.get(constants.CFG_KEY_HEADLESS_MODE) and appcfg.get(constants.CFG_KEY_HEADLESS_MODE).lower() == 'true':
        is_headless = True

    cfg_value = appcfg.get(constants.CFG_KEY_BROWSER_TYPE)
    if cfg_value and cfg_value.lower() == 'chrome':
        mylogger.debug('Chrome type browser return')
        return playwright.chromium.launch(headless=is_headless)
    elif cfg_value and cfg_value.lower() == 'firefox':
        mylogger.debug('Firefox type browser return')
        return playwright.firefox.launch(headless=is_headless)
    elif cfg_value and cfg_value.lower() == 'webkit':
        mylogger.debug('Webkit type browser return')
        return playwright.webkit.launch(headless=is_headless)

    mylogger.debug('No playwright return')
    return None

def navigate_to(page: Page, explicit_waittime: int, page_url: str, match_title: Union[Pattern[str], str], retry_count: int = 0) -> bool:
    mylogger = logging.getLogger(__name__)

    tryCount = 0
    while tryCount <= retry_count:
        tryCount += 1
        try:
            mylogger.debug(f'Loading URL - ({tryCount}:{retry_count}): {page_url}')
            page.goto(page_url, timeout=explicit_waittime * 1000)

            expect(page).to_have_title(match_title, timeout=explicit_waittime * 1000)

            if tryCount > 1:
                mylogger.info(f'Retry success at count: {tryCount}')
            return True
        except (AssertionError, TimeoutError):
            mylogger.error(f'Page title not match - ({tryCount}:{retry_count}) with "{match_title}":"{page.title()}"')

    return False

def check_title(page: Page, explicit_waittime: int, pattern: Union[Pattern[str], str]) -> bool:
    mylogger = logging.getLogger(__name__)
    try:
        expect(page).to_have_title(pattern, timeout=explicit_waittime * 1000)
        return True
    except AssertionError:
        mylogger.error(f'Page title not match with "{pattern}":"{page.title()}"')
    return False

def check_url(page: Page, explicit_waittime: int, pattern: Union[Pattern[str], str]) -> bool:
    mylogger = logging.getLogger(__name__)
    try:
        expect(page).to_have_url(pattern, timeout=explicit_waittime * 1000)
        return True
    except AssertionError:
        mylogger.error(f'Current URL does not match with "{pattern}":"{page.url}"')
    return False

def expect_to_be_attached(locator: Locator, explicit_waittime: int, dump_source_on_err: bool = True) -> bool:
    mylogger = logging.getLogger(__name__)
    try:
        expect(locator).to_be_attached(timeout=explicit_waittime * 1000)
        return True
    except AssertionError:
        mylogger.error('Cannot get element by locator')
        if dump_source_on_err:
            mylogger.debug(f'Page source dump: {locator.page.content()}')
    return False

def expect_to_be_visible(locator: Locator, explicit_waittime: int, dump_source_on_err: bool = True) -> bool:
    mylogger = logging.getLogger(__name__)
    try:
        expect(locator).to_be_visible(timeout=explicit_waittime * 1000)
        return True
    except AssertionError:
        mylogger.error('Element not visible')
        if dump_source_on_err:
            mylogger.debug(f'Page source dump: {locator.page.content()}')
    return False

def expect_to_have_count(locator: Locator, explicit_waittime: int, count: int, dump_source_on_err: bool = True) -> bool:
    mylogger = logging.getLogger(__name__)
    try:
        expect(locator).to_have_count(count, timeout=explicit_waittime * 1000)
        return True
    except AssertionError:
        mylogger.error('Element count not match by locator')
        if dump_source_on_err:
            mylogger.debug(f'Page source dump: {locator.page.content()}')
    return False



from typing import List, Optional
import logging
import re
import requests
from requests.exceptions import RequestException

TELEGRAM_BOT_SENDMSG_ENDPONT = 'https://api.telegram.org/bot%s/sendMessage'
TELEGRAM_BOT_SENDMSG_PARAM_CHATID = 'chat_id'
TELEGRAM_BOT_SENDMSG_PARAM_MSGTEXT = 'text'
TELEGRAM_BOT_SENDMSG_PARAM_PARSE_MODE = 'parse_mode'
TELEGRAM_BOT_SENDMSG_MAXLENGTH = 4096
TELEGRAM_BOT_SENDMSG_VALUE_PARSE_MODE_HTML = 'HTML'


class TelegramBot():
    def __init__(self, token: str, botname: Optional[str] = None) -> None:
        self.token = token
        self.botname = botname

        # Init logger for whole class
        self.mylogger = logging.getLogger(__name__)

    def postmsg(self, chatID_list: List[str], msg: str) -> int:
        posturl = TELEGRAM_BOT_SENDMSG_ENDPONT % self.token
        
        # Filter restrict char
        outgoing_msg: str = self.__filter(msg if self.botname is None else self.botname + ':' + msg)

        # Trim down length
        if len(outgoing_msg) > TELEGRAM_BOT_SENDMSG_MAXLENGTH:
            outgoing_msg = outgoing_msg[:TELEGRAM_BOT_SENDMSG_MAXLENGTH]
        
        self.mylogger.debug('Start postMsg URL: %s' % posturl)
        self.mylogger.debug('Outgoing msg: %s' % outgoing_msg)

        try:
            for chatID in chatID_list:
                self.mylogger.debug('Posting to chatID: %s' % chatID)

                # Define telegram api json payload
                json_payload = {}
                json_payload[TELEGRAM_BOT_SENDMSG_PARAM_CHATID] = chatID
                json_payload[TELEGRAM_BOT_SENDMSG_PARAM_MSGTEXT] = outgoing_msg
                json_payload[TELEGRAM_BOT_SENDMSG_PARAM_PARSE_MODE] = TELEGRAM_BOT_SENDMSG_VALUE_PARSE_MODE_HTML
                self.mylogger.debug('outputJson: %s' % json_payload)

                response = requests.post(posturl, json=json_payload)
                if response.status_code != requests.codes.ok:
                    self.mylogger.debug('PostMsg error with chatID: %s - response: %s' % (chatID, response.text))
                    return -1
                
                json_response = response.json()
                if not(json_response.get('ok')) or json_response.get('ok') != True:
                    self.mylogger.debug('PostMsg error with chatID: %s - response: %s' % (chatID, response.json()))
                    return -1
                
                self.mylogger.debug('PostMsg done with chatID: %s' % chatID)
                return 0
        except RequestException:
            self.mylogger.exception('Error during postMsg')

        return -2
    
    # Some char cannot be send via tg, so need convert into html entity
    def __filter(self, sourceMsg: str) -> str:
        returnMsg = re.sub(r'&', '&amp;', sourceMsg)
        returnMsg = re.sub(r'<', '&lt;', returnMsg)
        returnMsg = re.sub(r'>', '&gt;', returnMsg)
        return returnMsg
    



import logging
from typing import Optional, List, Union
from selenium.webdriver.remote.webdriver import WebDriver as RemoteWebDriver
from selenium.webdriver.remote.webdriver import WebElement
from selenium.webdriver.common.alert import Alert
from selenium.webdriver.common.by import By
from selenium.webdriver.support.wait import WebDriverWait
from selenium.webdriver.support import expected_conditions as ExpectedConditions
from selenium.common.exceptions import TimeoutException
from selenium.common.exceptions import NoSuchElementException
from selenium.common.exceptions import NoAlertPresentException

# normal webdriver
from selenium import webdriver

from ..cfgloader import AppCfg
from . import constants

#general method for getting webdriver
def get_webdriver(appcfg: AppCfg) -> Optional[RemoteWebDriver]:
   mylogger = logging.getLogger(__name__)

   is_headless = False
   if appcfg.get(constants.CFG_KEY_HEADLESS_MODE) and appcfg.get(constants.CFG_KEY_HEADLESS_MODE).lower() == 'true':
      is_headless = True

   cfg_value = appcfg.get(constants.CFG_KEY_BROWSER_TYPE)
   if cfg_value and cfg_value.lower() == 'chrome':
      chrome_option = webdriver.ChromeOptions()
      
      if appcfg.get(constants.CFG_KEY_BROWSER_BINARY):
         chrome_option.binary_location = appcfg.get(constants.CFG_KEY_BROWSER_BINARY)

      if is_headless:
         chrome_option.add_argument('--headless=new')
         #override user-agent as some site disallow headless mode browser
         chrome_option.add_argument('--user-agent=' + constants.DEFAULT_USER_AGENT_CHROME)
      
      mylogger.debug('Chrome type driver return')
      return webdriver.Chrome(options=chrome_option)
   elif cfg_value and cfg_value.lower() == 'firefox':
      firefox_option = webdriver.FirefoxOptions()
      if appcfg.get(constants.CFG_KEY_BROWSER_BINARY):
         firefox_option.binary_location = appcfg.get(constants.CFG_KEY_BROWSER_BINARY)
      
      if is_headless:
         firefox_option.add_argument('-headless')

      mylogger.debug('Firefox type driver return')
      return webdriver.Firefox(options=firefox_option)
   
   mylogger.debug('No webdriver')
   return None

def navigate_to(remote_webdriver: RemoteWebDriver, explicit_waittime: int, page_url: str, match_title: str, retry_count: int = 0, partial_match_only: bool = False) -> bool:
    mylogger = logging.getLogger(__name__)

    tryCount = 0
    while tryCount <= retry_count:
        tryCount += 1
        try:
            mylogger.debug('Loading URL - (%d:%d): %s' % (tryCount, retry_count, page_url))
            remote_webdriver.get(page_url)

            WebDriverWait(remote_webdriver, explicit_waittime).until(
                ExpectedConditions.title_is(match_title) \
                    if not partial_match_only else ExpectedConditions.title_contains(match_title)
            )
            if tryCount > 1:
                mylogger.info('Retry success at count: %d' % tryCount)
            return True
        except TimeoutException:
            mylogger.error('Page title not match - (%d:%d) with "%s":"%s"' % (tryCount, retry_count, match_title, remote_webdriver.title))

    return False

def check_url(remote_webdriver: RemoteWebDriver, explicit_waittime: int, pattern: str) -> bool:
    mylogger = logging.getLogger(__name__)
    try:
        WebDriverWait(remote_webdriver, explicit_waittime).until(
            ExpectedConditions.url_matches(pattern)
        )
        return True
    except TimeoutException:
        mylogger.error('Current URL does not match with "%s":"%s"' % (pattern, remote_webdriver.current_url))
    return False

def get_element_by_css(remote_webdriver: RemoteWebDriver, explicit_waittime: int, css_selector: str) -> Optional[WebElement]:
    mylogger = logging.getLogger(__name__)
    try:
        return WebDriverWait(remote_webdriver, explicit_waittime).until(
            ExpectedConditions.presence_of_element_located((By.CSS_SELECTOR, css_selector))
        )
    except TimeoutException:
        mylogger.error('Cannot get element by css: %s' % css_selector)
        mylogger.debug('Page source dump: %s' % remote_webdriver.page_source)
    return None

def find_element_by_css(remote_webdriver: RemoteWebDriver, css_selector: str) -> Optional[WebElement]:
    """No wait and no error log method, just get the element by css
    """
    try:
        return remote_webdriver.find_element(by=By.CSS_SELECTOR, value=css_selector)
    except NoSuchElementException:
        return None

def get_nested_element_by_css(remote_webdriver: RemoteWebDriver, explicit_waittime: int, parent_element: WebElement, css_selector: str) -> Optional[WebElement]:
    mylogger = logging.getLogger(__name__)
    try:
        return WebDriverWait(parent_element, explicit_waittime).until(
            ExpectedConditions.presence_of_element_located((By.CSS_SELECTOR, css_selector))
        )
    except TimeoutException:
        mylogger.error('Cannot get nested element by css: %s' % css_selector)
        mylogger.debug('Page source dump: %s' % remote_webdriver.page_source)
    return None

def find_nested_element_by_css(parent_element: WebElement, css_selector: str) -> Optional[WebElement]:
    """No wait and no error log method, just get the nested element by css
    """
    try:
        return parent_element.find_element(by=By.CSS_SELECTOR, value=css_selector)
    except NoSuchElementException:
        return None

def get_elements_by_css(remote_webdriver: RemoteWebDriver, explicit_waittime: int, css_selector: str) -> Optional[List[WebElement]]:
    mylogger = logging.getLogger(__name__)
    try:
        return WebDriverWait(remote_webdriver, explicit_waittime).until(
            ExpectedConditions.presence_of_all_elements_located((By.CSS_SELECTOR, css_selector))
        )
    except TimeoutException:
        mylogger.error('Cannot get any element by css: %s' % css_selector)
        mylogger.debug('Page source dump: %s' % remote_webdriver.page_source)
    return None

def find_elements_by_css(remote_webdriver: RemoteWebDriver, css_selector: str) -> Optional[List[WebElement]]:
    """No wait and no error log method, just get list of elements by css
    """
    try:
        return remote_webdriver.find_elements(by=By.CSS_SELECTOR, value=css_selector)
    except NoSuchElementException:
        return None

def get_nested_elements_by_css(remote_webdriver: RemoteWebDriver, explicit_waittime: int, parent_element: WebElement, css_selector: str) -> Optional[List[WebElement]]:
    mylogger = logging.getLogger(__name__)
    try:
        return WebDriverWait(parent_element, explicit_waittime).until(
            ExpectedConditions.presence_of_all_elements_located((By.CSS_SELECTOR, css_selector))
        )
    except TimeoutException:
        mylogger.error('Cannot get any nested element by css: %s' % css_selector)
        mylogger.debug('Page source dump: %s' % remote_webdriver.page_source)
    return None

def find_nested_elements_by_css(parent_element: WebElement, css_selector: str) -> Optional[List[WebElement]]:
    """No wait and no error log method, just get list of nested elements by css
    """
    try:
        return parent_element.find_elements(by=By.CSS_SELECTOR, value=css_selector)
    except NoSuchElementException:
        return None
    
def get_pop_alert(remote_webdriver: RemoteWebDriver, explicit_waittime: int) -> Optional[Alert]:
    mylogger = logging.getLogger(__name__)
    try:
        return WebDriverWait(remote_webdriver, explicit_waittime).until(
            ExpectedConditions.alert_is_present()
        )
    except TimeoutException:
        mylogger.error('No alert at all')
        mylogger.debug('Page url: %s' % remote_webdriver.current_url)
    return None

def find_pop_alert(remote_webdriver: RemoteWebDriver) -> Optional[Alert]:
    try:
        return remote_webdriver.switch_to.alert
    except NoAlertPresentException:
        return None

def check_text_exist(remote_webdriver: RemoteWebDriver, explicit_waittime: int, css_selector: str, text: str) -> bool:
    mylogger = logging.getLogger(__name__)
    try:
        return WebDriverWait(remote_webdriver, explicit_waittime).until(
            ExpectedConditions.text_to_be_present_in_element((By.CSS_SELECTOR, css_selector), text)
        )
    except TimeoutException:
        mylogger.error('Text not exist by css: %s - %s' % (css_selector, text))
        mylogger.debug('Page source dump: %s' % remote_webdriver.page_source)
    return False
